@@ -26,6 +26,7 @@
 #include "talloc.h"
 
 #include "common/msg.h"
+#include "common/encode.h"
 #include "options/options.h"
 #include "common/common.h"
 
@@ -39,11 +40,11 @@
 #include "video/decode/dec_video.h"
 
 #include "core.h"
+#include "command.h"
 
 static int build_afilter_chain(struct MPContext *mpctx)
 {
     struct dec_audio *d_audio = mpctx->d_audio;
-    struct ao *ao = mpctx->ao;
     struct MPOpts *opts = mpctx->opts;
 
     if (!d_audio)
@@ -52,23 +53,20 @@ static int build_afilter_chain(struct MPContext *mpctx)
     struct mp_audio in_format;
     mp_audio_buffer_get_format(d_audio->decode_buffer, &in_format);
 
+    struct mp_audio out_format;
+    ao_get_format(mpctx->ao, &out_format);
+
     int new_srate;
     if (af_control_any_rev(d_audio->afilter, AF_CONTROL_SET_PLAYBACK_SPEED,
                            &opts->playback_speed))
         new_srate = in_format.rate;
     else {
         new_srate = in_format.rate * opts->playback_speed;
-        if (new_srate != ao->samplerate) {
-            // limits are taken from libaf/af_resample.c
-            if (new_srate < 8000)
-                new_srate = 8000;
-            if (new_srate > 192000)
-                new_srate = 192000;
+        if (new_srate != out_format.rate)
             opts->playback_speed = new_srate / (double)in_format.rate;
-        }
     }
     return audio_init_filters(d_audio, new_srate,
-                              &ao->samplerate, &ao->channels, &ao->format);
+                &out_format.rate, &out_format.channels, &out_format.format);
 }
 
 static int recreate_audio_filters(struct MPContext *mpctx)
@@ -90,7 +88,7 @@ int reinit_audio_filters(struct MPContext *mpctx)
 {
     struct dec_audio *d_audio = mpctx->d_audio;
     if (!d_audio)
-        return -2;
+        return 0;
 
     af_uninit(mpctx->d_audio->afilter);
     if (af_init(mpctx->d_audio->afilter) < 0)
@@ -98,7 +96,7 @@ int reinit_audio_filters(struct MPContext *mpctx)
     if (recreate_audio_filters(mpctx) < 0)
         return -1;
 
-    return 0;
+    return 1;
 }
 
 void reinit_audio_chain(struct MPContext *mpctx)
@@ -111,6 +109,8 @@ void reinit_audio_chain(struct MPContext *mpctx)
         goto no_audio;
     }
 
+    mp_notify(mpctx, MPV_EVENT_AUDIO_RECONFIG, NULL);
+
     if (!(mpctx->initialized_flags & INITIALIZED_ACODEC)) {
         mpctx->initialized_flags |= INITIALIZED_ACODEC;
         assert(!mpctx->d_audio);
@@ -119,6 +119,8 @@ void reinit_audio_chain(struct MPContext *mpctx)
         mpctx->d_audio->global = mpctx->global;
         mpctx->d_audio->opts = opts;
         mpctx->d_audio->header = sh;
+        mpctx->d_audio->metadata = mpctx->demuxer->metadata;
+        mpctx->d_audio->replaygain_data = mpctx->demuxer->replaygain_data;
         if (!audio_init_best_codec(mpctx->d_audio, opts->audio_decoders))
             goto init_error;
     }
@@ -127,20 +129,25 @@ void reinit_audio_chain(struct MPContext *mpctx)
     struct mp_audio in_format;
     mp_audio_buffer_get_format(mpctx->d_audio->decode_buffer, &in_format);
 
+    if (mpctx->ao_decoder_fmt && (mpctx->initialized_flags & INITIALIZED_AO) &&
+        !mp_audio_config_equals(mpctx->ao_decoder_fmt, &in_format) &&
+        opts->gapless_audio < 0)
+    {
+        uninit_player(mpctx, INITIALIZED_AO);
+    }
+
     int ao_srate = opts->force_srate;
     int ao_format = opts->audio_output_format;
     struct mp_chmap ao_channels = {0};
     if (mpctx->initialized_flags & INITIALIZED_AO) {
-        ao_srate    = mpctx->ao->samplerate;
-        ao_format   = mpctx->ao->format;
-        ao_channels = mpctx->ao->channels;
+        struct mp_audio out_format;
+        ao_get_format(mpctx->ao, &out_format);
+        ao_srate    = out_format.rate;
+        ao_format   = out_format.format;
+        ao_channels = out_format.channels;
     } else {
-        // Automatic downmix
-        if (mp_chmap_is_stereo(&opts->audio_output_channels) &&
-            !mp_chmap_is_stereo(&in_format.channels))
-        {
-            mp_chmap_from_channels(&ao_channels, 2);
-        }
+        if (!AF_FORMAT_IS_SPECIAL(in_format.format))
+            ao_channels = opts->audio_output_channels; // automatic downmix
     }
 
     // Determine what the filter chain outputs. build_afilter_chain() also
@@ -168,14 +175,19 @@ void reinit_audio_chain(struct MPContext *mpctx)
             goto init_error;
         }
 
-        ao->buffer = mp_audio_buffer_create(ao);
-        mp_audio_buffer_reinit_fmt(ao->buffer, ao->format, &ao->channels,
-                                   ao->samplerate);
+        struct mp_audio fmt;
+        ao_get_format(ao, &fmt);
 
-        char *s = mp_audio_fmt_to_str(ao->samplerate, &ao->channels, ao->format);
-        MP_INFO(mpctx, "AO: [%s] %s\n", ao->driver->name, s);
+        mpctx->ao_buffer = mp_audio_buffer_create(ao);
+        mp_audio_buffer_reinit(mpctx->ao_buffer, &fmt);
+
+        mpctx->ao_decoder_fmt = talloc(NULL, struct mp_audio);
+        *mpctx->ao_decoder_fmt = in_format;
+
+        char *s = mp_audio_config_to_str(&fmt);
+        MP_INFO(mpctx, "AO: [%s] %s\n", ao_get_name(ao), s);
         talloc_free(s);
-        MP_VERBOSE(mpctx, "AO: Description: %s\n", ao->driver->description);
+        MP_VERBOSE(mpctx, "AO: Description: %s\n", ao_get_description(ao));
         update_window_title(mpctx, true);
     }
 
@@ -224,7 +236,7 @@ double written_audio_pts(struct MPContext *mpctx)
 
     // Data that was ready for ao but was buffered because ao didn't fully
     // accept everything to internal buffers yet
-    buffered_output += mp_audio_buffer_seconds(mpctx->ao->buffer);
+    buffered_output += mp_audio_buffer_seconds(mpctx->ao_buffer);
 
     // Filters divide audio length by playback_speed, so multiply by it
     // to get the length in original units without speedup or slowdown
@@ -248,8 +260,15 @@ static int write_to_ao(struct MPContext *mpctx, struct mp_audio *data, int flags
     if (mpctx->paused)
         return 0;
     struct ao *ao = mpctx->ao;
-    ao->pts = pts;
-    double real_samplerate = ao->samplerate / mpctx->opts->playback_speed;
+    struct mp_audio out_format;
+    ao_get_format(ao, &out_format);
+    mpctx->ao_pts = pts;
+#if HAVE_ENCODING
+    encode_lavc_set_audio_pts(mpctx->encode_lavc_ctx, mpctx->ao_pts);
+#endif
+    if (data->samples == 0)
+        return 0;
+    double real_samplerate = out_format.rate / mpctx->opts->playback_speed;
     int played = ao_play(mpctx->ao, data->planes, data->samples, flags);
     assert(played <= data->samples);
     if (played > 0) {
@@ -257,7 +276,7 @@ static int write_to_ao(struct MPContext *mpctx, struct mp_audio *data, int flags
         mpctx->delay += played / real_samplerate;
         // Keep correct pts for remaining data - could be used to flush
         // remaining buffer when closing ao.
-        ao->pts += played / real_samplerate;
+        mpctx->ao_pts += played / real_samplerate;
         return played;
     }
     return 0;
@@ -267,7 +286,7 @@ static int write_silence_to_ao(struct MPContext *mpctx, int samples, int flags,
                                double pts)
 {
     struct mp_audio tmp = {0};
-    mp_audio_buffer_get_format(mpctx->ao->buffer, &tmp);
+    mp_audio_buffer_get_format(mpctx->ao_buffer, &tmp);
     tmp.samples = samples;
     char *p = talloc_size(NULL, tmp.samples * tmp.sstride);
     for (int n = 0; n < tmp.num_planes; n++)
@@ -288,15 +307,18 @@ static int audio_start_sync(struct MPContext *mpctx, int playsize)
 
     assert(d_audio);
 
+    struct mp_audio out_format;
+    ao_get_format(ao, &out_format);
+
     // Timing info may not be set without
-    res = audio_decode(d_audio, ao->buffer, 1);
+    res = audio_decode(d_audio, mpctx->ao_buffer, 1);
     if (res < 0)
         return res;
 
     int samples;
     bool did_retry = false;
     double written_pts;
-    double real_samplerate = ao->samplerate / opts->playback_speed;
+    double real_samplerate = out_format.rate / opts->playback_speed;
     bool hrseek = mpctx->hrseek_active;   // audio only hrseek
     mpctx->hrseek_active = false;
     while (1) {
@@ -306,14 +328,14 @@ static int audio_start_sync(struct MPContext *mpctx, int playsize)
             ptsdiff = written_pts - mpctx->hrseek_pts;
         else
             ptsdiff = written_pts - mpctx->video_next_pts - mpctx->delay
-                      - mpctx->audio_delay;
+                      + mpctx->audio_delay;
         samples = ptsdiff * real_samplerate;
 
         // ogg demuxers give packets without timing
         if (written_pts <= 1 && d_audio->pts == MP_NOPTS_VALUE) {
             if (!did_retry) {
                 // Try to read more data to see packets that have pts
-                res = audio_decode(d_audio, ao->buffer, ao->samplerate);
+                res = audio_decode(d_audio, mpctx->ao_buffer, out_format.rate);
                 if (res < 0)
                     return res;
                 did_retry = true;
@@ -331,16 +353,14 @@ static int audio_start_sync(struct MPContext *mpctx, int playsize)
         mpctx->syncing_audio = false;
         int skip_samples = -samples;
         int a = MPMIN(skip_samples, MPMAX(playsize, 2500));
-        res = audio_decode(d_audio, ao->buffer, a);
-        if (skip_samples <= mp_audio_buffer_samples(ao->buffer)) {
-            mp_audio_buffer_skip(ao->buffer, skip_samples);
-            ao->buffer_playable_samples = 0;
+        res = audio_decode(d_audio, mpctx->ao_buffer, a);
+        if (skip_samples <= mp_audio_buffer_samples(mpctx->ao_buffer)) {
+            mp_audio_buffer_skip(mpctx->ao_buffer, skip_samples);
             if (res < 0)
                 return res;
-            return audio_decode(d_audio, ao->buffer, playsize);
+            return audio_decode(d_audio, mpctx->ao_buffer, playsize);
         }
-        mp_audio_buffer_clear(ao->buffer);
-        ao->buffer_playable_samples = 0;
+        mp_audio_buffer_clear(mpctx->ao_buffer);
         if (res < 0)
             return res;
     }
@@ -350,15 +370,15 @@ static int audio_start_sync(struct MPContext *mpctx, int playsize)
     if (samples >= playsize) {
         /* This case could fall back to the one below with
          * samples = playsize, but then silence would keep accumulating
-         * in ao->buffer if the AO accepts less data than it asks for
+         * in ao_buffer if the AO accepts less data than it asks for
          * in playsize. */
         write_silence_to_ao(mpctx, playsize, 0,
                             written_pts - samples / real_samplerate);
         return ASYNC_PLAY_DONE;
     }
     mpctx->syncing_audio = false;
-    mp_audio_buffer_prepend_silence(ao->buffer, samples);
-    return audio_decode(d_audio, ao->buffer, playsize);
+    mp_audio_buffer_prepend_silence(mpctx->ao_buffer, samples);
+    return audio_decode(d_audio, mpctx->ao_buffer, playsize);
 }
 
 int fill_audio_out_buffers(struct MPContext *mpctx, double endpts)
@@ -371,8 +391,10 @@ int fill_audio_out_buffers(struct MPContext *mpctx, double endpts)
     bool signal_eof = false;
     bool partial_fill = false;
     struct dec_audio *d_audio = mpctx->d_audio;
+    struct mp_audio out_format;
+    ao_get_format(ao, &out_format);
     // Can't adjust the start of audio with spdif pass-through.
-    bool modifiable_audio_format = !(ao->format & AF_FORMAT_SPECIAL_MASK);
+    bool modifiable_audio_format = !(out_format.format & AF_FORMAT_SPECIAL_MASK);
 
     assert(d_audio);
 
@@ -393,7 +415,7 @@ int fill_audio_out_buffers(struct MPContext *mpctx, double endpts)
     if (mpctx->syncing_audio || mpctx->hrseek_active)
         res = audio_start_sync(mpctx, playsize);
     else
-        res = audio_decode(d_audio, ao->buffer, playsize);
+        res = audio_decode(d_audio, mpctx->ao_buffer, playsize);
 
     if (res < 0) {  // EOF, error or format change
         if (res == -2) {
@@ -401,7 +423,7 @@ int fill_audio_out_buffers(struct MPContext *mpctx, double endpts)
              * implementation would require draining buffered old-format audio
              * while displaying video, then doing the output format switch.
              */
-            if (!mpctx->opts->gapless_audio)
+            if (mpctx->opts->gapless_audio < 1)
                 uninit_player(mpctx, INITIALIZED_AO);
             reinit_audio_chain(mpctx);
             return -1;
@@ -412,8 +434,8 @@ int fill_audio_out_buffers(struct MPContext *mpctx, double endpts)
     }
 
     if (endpts != MP_NOPTS_VALUE) {
-        double samples = (endpts - written_audio_pts(mpctx) + mpctx->audio_delay)
-                         * ao->samplerate / opts->playback_speed;
+        double samples = (endpts - written_audio_pts(mpctx) - mpctx->audio_delay)
+                         * out_format.rate / opts->playback_speed;
         if (playsize > samples) {
             playsize = MPMAX(samples, 0);
             audio_eof = true;
@@ -421,8 +443,8 @@ int fill_audio_out_buffers(struct MPContext *mpctx, double endpts)
         }
     }
 
-    if (playsize > mp_audio_buffer_samples(ao->buffer)) {
-        playsize = mp_audio_buffer_samples(ao->buffer);
+    if (playsize > mp_audio_buffer_samples(mpctx->ao_buffer)) {
+        playsize = mp_audio_buffer_samples(mpctx->ao_buffer);
         partial_fill = true;
     }
     if (!playsize)
@@ -438,22 +460,16 @@ int fill_audio_out_buffers(struct MPContext *mpctx, double endpts)
         }
     }
 
-    assert(ao->buffer_playable_samples <= mp_audio_buffer_samples(ao->buffer));
+    if (mpctx->paused)
+        playsize = 0;
 
     struct mp_audio data;
-    mp_audio_buffer_peek(ao->buffer, &data);
+    mp_audio_buffer_peek(mpctx->ao_buffer, &data);
     data.samples = MPMIN(data.samples, playsize);
     int played = write_to_ao(mpctx, &data, playflags, written_audio_pts(mpctx));
     assert(played >= 0 && played <= data.samples);
-    ao->buffer_playable_samples = playsize - played;
 
-    if (played > 0) {
-        mp_audio_buffer_skip(ao->buffer, played);
-    } else if (!mpctx->paused && audio_eof && ao_get_delay(ao) < .04) {
-        // Sanity check to avoid hanging in case current ao doesn't output
-        // partial chunks and doesn't check for AOPLAY_FINAL_CHUNK
-        signal_eof = true;
-    }
+    mp_audio_buffer_skip(mpctx->ao_buffer, played);
 
     return signal_eof ? -2 : -partial_fill;
 }
@@ -463,8 +479,7 @@ void clear_audio_output_buffers(struct MPContext *mpctx)
 {
     if (mpctx->ao) {
         ao_reset(mpctx->ao);
-        mp_audio_buffer_clear(mpctx->ao->buffer);
-        mpctx->ao->buffer_playable_samples = 0;
+        mp_audio_buffer_clear(mpctx->ao_buffer);
     }
 }
 
