@@ -23,29 +23,29 @@
 #include <assert.h>
 #include <ctype.h>
 #include <string.h>
+#include <pthread.h>
 
 #include "config.h"
 #include "talloc.h"
 
+#include "misc/dispatch.h"
 #include "osdep/io.h"
-#include "osdep/priority.h"
 #include "osdep/terminal.h"
 #include "osdep/timer.h"
 
 #include "common/av_log.h"
 #include "common/codecs.h"
-#include "common/cpudetect.h"
 #include "common/encode.h"
 #include "options/m_config.h"
 #include "options/m_option.h"
 #include "options/m_property.h"
 #include "common/common.h"
 #include "common/msg.h"
+#include "common/msg_control.h"
 #include "common/global.h"
 #include "options/parse_configfile.h"
 #include "options/parse_commandline.h"
 #include "common/playlist.h"
-#include "common/playlist_parser.h"
 #include "options/options.h"
 #include "input/input.h"
 
@@ -60,7 +60,7 @@
 #include "video/out/vo.h"
 
 #include "core.h"
-#include "lua.h"
+#include "client.h"
 #include "command.h"
 #include "screenshot.h"
 
@@ -72,10 +72,6 @@
 #endif
 #endif
 
-#if HAVE_X11
-#include "video/out/x11_common.h"
-#endif
-
 #if HAVE_COCOA
 #include "osdep/macosx_application.h"
 #endif
@@ -84,6 +80,8 @@
 #include <pthread.h>
 #endif
 
+static bool terminal_initialized;
+
 const char mp_help_text[] =
 "Usage:   mpv [options] [url|path/]filename\n"
 "\n"
@@ -92,7 +90,7 @@ const char mp_help_text[] =
 " --no-audio        do not play sound\n"
 " --no-video        do not play video\n"
 " --fs              fullscreen playback\n"
-" --sub=<file>      specify subtitle file to use\n"
+" --sub-file=<file> specify subtitle file to use\n"
 " --playlist=<file> specify playlist file\n"
 "\n"
 " --list-options    list all mpv options\n"
@@ -102,17 +100,26 @@ void mp_print_version(struct mp_log *log, int always)
 {
     int v = always ? MSGL_INFO : MSGL_V;
     mp_msg(log, v,
-           "%s (C) 2000-2013 mpv/MPlayer/mplayer2 projects\n built on %s\n",
+           "%s (C) 2000-2014 mpv/MPlayer/mplayer2 projects\n built on %s\n",
            mpv_version, mpv_builddate);
     print_libav_versions(log, v);
     mp_msg(log, v, "\n");
 }
 
-static MP_NORETURN void exit_player(struct MPContext *mpctx,
-                                    enum exit_reason how)
+static void shutdown_clients(struct MPContext *mpctx)
 {
-    int rc;
-    uninit_player(mpctx, INITIALIZED_ALL);
+    while (mpctx->clients && mp_clients_num(mpctx)) {
+        mp_client_broadcast_event(mpctx, MPV_EVENT_SHUTDOWN, NULL);
+        mp_dispatch_queue_process(mpctx->dispatch, 0);
+        mp_input_get_cmd(mpctx->input, 100, 1);
+    }
+    mp_clients_destroy(mpctx);
+}
+
+void mp_destroy(struct MPContext *mpctx)
+{
+    if (mpctx->initialized)
+        uninit_player(mpctx, INITIALIZED_ALL);
 
 #if HAVE_ENCODING
     encode_lavc_finish(mpctx->encode_lavc_ctx);
@@ -121,32 +128,42 @@ static MP_NORETURN void exit_player(struct MPContext *mpctx,
 
     mpctx->encode_lavc_ctx = NULL;
 
-#if HAVE_LUA
-    mp_lua_uninit(mpctx);
-#endif
-
-#if defined(__MINGW32__)
-    timeEndPeriod(1);
-#endif
-
-#if HAVE_COCOA
-    cocoa_set_input_context(NULL);
-#endif
+    shutdown_clients(mpctx);
 
     command_uninit(mpctx);
 
+    mp_dispatch_set_wakeup_fn(mpctx->dispatch, NULL, NULL);
     mp_input_uninit(mpctx->input);
 
     osd_free(mpctx->osd);
 
 #if HAVE_LIBASS
-    ass_library_done(mpctx->ass_library);
-    mpctx->ass_library = NULL;
+    if (mpctx->ass_library)
+        ass_library_done(mpctx->ass_library);
 #endif
 
-    getch2_disable();
+    if (mpctx->opts->use_terminal) {
+        getch2_disable();
+        terminal_initialized = false;
+    }
     uninit_libav(mpctx->global);
 
+    if (mpctx->autodetach)
+        pthread_detach(pthread_self());
+
+    mp_msg_uninit(mpctx->global);
+    talloc_free(mpctx);
+}
+
+static MP_NORETURN void exit_player(struct MPContext *mpctx,
+                                    enum exit_reason how)
+{
+    int rc;
+
+#if HAVE_COCOA
+    cocoa_set_input_context(NULL);
+#endif
+
     if (how != EXIT_NONE) {
         const char *reason;
         switch (how) {
@@ -181,11 +198,7 @@ static MP_NORETURN void exit_player(struct MPContext *mpctx,
         }
     }
 
-    // must be last since e.g. mp_msg uses option values
-    // that will be freed by this.
-
-    mp_msg_uninit(mpctx->global);
-    talloc_free(mpctx);
+    mp_destroy(mpctx);
 
 #if HAVE_COCOA
     terminate_cocoa_application();
@@ -214,13 +227,6 @@ static bool handle_help_options(struct MPContext *mpctx)
         talloc_free(list);
         opt_exit = 1;
     }
-#if HAVE_X11
-    if (opts->vo.fstype_list && strcmp(opts->vo.fstype_list[0], "help") == 0) {
-        fstype_help(log);
-        MP_INFO(mpctx, "\n");
-        opt_exit = 1;
-    }
-#endif
     if ((opts->demuxer_name && strcmp(opts->demuxer_name, "help") == 0) ||
         (opts->audio_demuxer_name && strcmp(opts->audio_demuxer_name, "help") == 0) ||
         (opts->sub_demuxer_name && strcmp(opts->sub_demuxer_name, "help") == 0)) {
@@ -229,7 +235,7 @@ static bool handle_help_options(struct MPContext *mpctx)
         opt_exit = 1;
     }
 #if HAVE_ENCODING
-    if (encode_lavc_showhelp(log, &opts->encode_output))
+    if (encode_lavc_showhelp(log, opts->encode_opts))
         opt_exit = 1;
 #endif
     return opt_exit;
@@ -282,32 +288,25 @@ static void osdep_preinit(int *p_argc, char ***p_argv)
     if (pSetSearchPathMode)
         pSetSearchPathMode(BASE_SEARCH_PATH_ENABLE_SAFE_SEARCHMODE);
 #endif
-
-    terminal_init();
-
-    mp_time_init();
 }
 
 static int cfg_include(void *ctx, char *filename, int flags)
 {
     struct MPContext *mpctx = ctx;
-    return m_config_parse_config_file(mpctx->mconfig, filename, flags);
+    return m_config_parse_config_file(mpctx->mconfig, filename, NULL, flags);
 }
 
-static int mpv_main(int argc, char *argv[])
+struct MPContext *mp_create(void)
 {
-    osdep_preinit(&argc, &argv);
-
-    if (argc >= 1) {
-        argc--;
-        argv++;
-    }
+    mp_time_init();
 
     struct MPContext *mpctx = talloc(NULL, MPContext);
     *mpctx = (struct MPContext){
-        .last_dvb_step = 1,
-        .terminal_osd_text = talloc_strdup(mpctx, ""),
+        .last_chapter = -2,
+        .term_osd_contents = talloc_strdup(mpctx, ""),
+        .osd_progbar = { .type = -1 },
         .playlist = talloc_struct(mpctx, struct playlist, {0}),
+        .dispatch = mp_dispatch_create(mpctx),
     };
 
     mpctx->global = talloc_zero(mpctx, struct mpv_global);
@@ -330,22 +329,164 @@ static int mpv_main(int argc, char *argv[])
     mpctx->mconfig->use_profiles = true;
     mpctx->mconfig->is_toplevel = true;
 
+    mpctx->global->opts = mpctx->opts;
+
+    screenshot_init(mpctx);
+    mpctx->mixer = mixer_init(mpctx, mpctx->global);
+    command_init(mpctx);
+    init_libav(mpctx->global);
+    mp_clients_init(mpctx);
+
+    return mpctx;
+}
+
+static int check_stream_interrupt(void *ctx)
+{
+    struct MPContext *mpctx = ctx;
+    return mp_input_check_interrupt(mpctx->input);
+}
+
+static void wakeup_playloop(void *ctx)
+{
+    struct MPContext *mpctx = ctx;
+    mp_input_wakeup(mpctx->input);
+}
+
+// Finish mpctx initialization. This must be done after setting up all options.
+// Some of the initializations depend on the options, and can't be changed or
+// undone later.
+// cplayer: true if called by the command line player, false for client API
+// Returns: <0 on error, 0 on success.
+int mp_initialize(struct MPContext *mpctx)
+{
     struct MPOpts *opts = mpctx->opts;
-    mpctx->global->opts = opts;
+
+    assert(!mpctx->initialized);
+
+    if (opts->dump_stats && opts->dump_stats[0]) {
+        if (mp_msg_open_stats_file(mpctx->global, opts->dump_stats) < 0)
+            MP_ERR(mpctx, "Failed to open stats file '%s'\n", opts->dump_stats);
+    }
+    MP_STATS(mpctx, "start init");
+
+    if (mpctx->opts->use_terminal && !terminal_initialized) {
+        terminal_initialized = true;
+        terminal_init();
+        mp_msg_update_msglevels(mpctx->global);
+    }
+
+    mpctx->input = mp_input_init(mpctx->global);
+    mpctx->global->stream_interrupt_cb = check_stream_interrupt;
+    mpctx->global->stream_interrupt_cb_ctx = mpctx;
+
+    mp_dispatch_set_wakeup_fn(mpctx->dispatch, wakeup_playloop, mpctx);
+
+#if HAVE_ENCODING
+    if (opts->encode_opts->file && opts->encode_opts->file[0]) {
+        mpctx->encode_lavc_ctx = encode_lavc_init(opts->encode_opts,
+                                                  mpctx->global);
+        if(!mpctx->encode_lavc_ctx) {
+            MP_INFO(mpctx, "Encoding initialization failed.");
+            return -1;
+        }
+        m_config_set_option0(mpctx->mconfig, "vo", "lavc");
+        m_config_set_option0(mpctx->mconfig, "ao", "lavc");
+        m_config_set_option0(mpctx->mconfig, "fixed-vo", "yes");
+        m_config_set_option0(mpctx->mconfig, "keep-open", "no");
+        m_config_set_option0(mpctx->mconfig, "force-window", "no");
+        m_config_set_option0(mpctx->mconfig, "gapless-audio", "yes");
+        m_config_set_option0(mpctx->mconfig, "resume-playback", "no");
+        m_config_set_option0(mpctx->mconfig, "load-scripts", "no");
+        m_config_set_option0(mpctx->mconfig, "osc", "no");
+        m_config_set_option0(mpctx->mconfig, "framedrop", "no");
+        mp_input_enable_section(mpctx->input, "encode", MP_INPUT_EXCLUSIVE);
+    }
+#endif
+
+    if (opts->use_terminal) {
+        if (mpctx->opts->slave_mode)
+            terminal_setup_stdin_cmd_input(mpctx->input);
+        else if (mpctx->opts->consolecontrols)
+            terminal_setup_getch(mpctx->input);
+
+        if (opts->consolecontrols)
+            getch2_enable();
+    }
+
+#if HAVE_LIBASS
+    mpctx->ass_log = mp_log_new(mpctx, mpctx->global->log, "!libass");
+    mpctx->ass_library = mp_ass_init(mpctx->global, mpctx->ass_log);
+#else
+    MP_WARN(mpctx, "Compiled without libass.\n");
+    MP_WARN(mpctx, "There will be no OSD and no text subtitles.\n");
+#endif
+
+    mpctx->osd = osd_create(mpctx->global);
+
+    // From this point on, all mpctx members are initialized.
+    mpctx->initialized = true;
+
+    mp_get_resume_defaults(mpctx);
+
+#if HAVE_COCOA
+    if (mpctx->is_cplayer)
+        cocoa_set_input_context(mpctx->input);
+#endif
+
+    if (opts->force_vo) {
+        opts->fixed_vo = 1;
+        mpctx->video_out = init_best_video_out(mpctx->global, mpctx->input,
+                                               mpctx->osd,
+                                               mpctx->encode_lavc_ctx);
+        if (!mpctx->video_out) {
+            MP_FATAL(mpctx, "Error opening/initializing "
+                    "the selected video_out (-vo) device.\n");
+            return -1;
+        }
+        mpctx->mouse_cursor_visible = true;
+        mpctx->initialized_flags |= INITIALIZED_VO;
+    }
+
+    // Lua user scripts (etc.) can call arbitrary functions. Load them at a point
+    // where this is safe.
+    mp_load_scripts(mpctx);
+
+    if (opts->shuffle)
+        playlist_shuffle(mpctx->playlist);
+
+    if (opts->merge_files)
+        merge_playlist_files(mpctx->playlist);
+
+    mpctx->playlist->current = mp_check_playlist_resume(mpctx, mpctx->playlist);
+    if (!mpctx->playlist->current)
+        mpctx->playlist->current = mpctx->playlist->first;
+
+    MP_STATS(mpctx, "end init");
+
+    return 0;
+}
+
+int mpv_main(int argc, char *argv[])
+{
+    osdep_preinit(&argc, &argv);
+
+    struct MPContext *mpctx = mp_create();
+    struct MPOpts *opts = mpctx->opts;
+
+    mpctx->is_cplayer = true;
 
     char *verbose_env = getenv("MPV_VERBOSE");
     if (verbose_env)
         opts->verbose = atoi(verbose_env);
-    mp_msg_update_msglevels(mpctx->global);
-
-    init_libav(mpctx->global);
-    GetCpuCaps(&gCpuCaps);
-    screenshot_init(mpctx);
-    mpctx->mixer = mixer_init(mpctx, mpctx->global);
-    command_init(mpctx);
 
     // Preparse the command line
     m_config_preparse_command_line(mpctx->mconfig, mpctx->global, argc, argv);
+
+    if (mpctx->opts->use_terminal && !terminal_initialized) {
+        terminal_initialized = true;
+        terminal_init();
+    }
+
     mp_msg_update_msglevels(mpctx->global);
 
     mp_print_version(mpctx->log, false);
@@ -381,80 +522,12 @@ static int mpv_main(int argc, char *argv[])
     }
 
 #if HAVE_PRIORITY
-    set_priority();
+    if (opts->w32_priority > 0)
+        SetPriorityClass(GetCurrentProcess(), opts->w32_priority);
 #endif
 
-    mpctx->input = mp_input_init(mpctx->global);
-    stream_set_interrupt_callback(mp_input_check_interrupt, mpctx->input);
-#if HAVE_COCOA
-    cocoa_set_input_context(mpctx->input);
-#endif
-
-#if HAVE_ENCODING
-    if (opts->encode_output.file && *opts->encode_output.file) {
-        mpctx->encode_lavc_ctx = encode_lavc_init(&opts->encode_output,
-                                                  mpctx->global);
-        if(!mpctx->encode_lavc_ctx) {
-            MP_INFO(mpctx, "Encoding initialization failed.");
-            exit_player(mpctx, EXIT_ERROR);
-        }
-        m_config_set_option0(mpctx->mconfig, "vo", "lavc");
-        m_config_set_option0(mpctx->mconfig, "ao", "lavc");
-        m_config_set_option0(mpctx->mconfig, "fixed-vo", "yes");
-        m_config_set_option0(mpctx->mconfig, "keep-open", "no");
-        m_config_set_option0(mpctx->mconfig, "force-window", "no");
-        m_config_set_option0(mpctx->mconfig, "gapless-audio", "yes");
-        m_config_set_option0(mpctx->mconfig, "resume-playback", "no");
-        mp_input_enable_section(mpctx->input, "encode", MP_INPUT_EXCLUSIVE);
-    }
-#endif
-
-    if (mpctx->opts->slave_mode)
-        terminal_setup_stdin_cmd_input(mpctx->input);
-    else if (mpctx->opts->consolecontrols)
-        terminal_setup_getch(mpctx->input);
-
-    if (opts->consolecontrols)
-        getch2_enable();
-
-#if HAVE_LIBASS
-    mpctx->ass_log = mp_log_new(mpctx, mpctx->global->log, "!libass");
-    mpctx->ass_library = mp_ass_init(mpctx->global, mpctx->ass_log);
-#else
-    MP_WARN(mpctx, "Compiled without libass.\n");
-    MP_WARN(mpctx, "There will be no OSD and no text subtitles.\n");
-#endif
-
-    mpctx->osd = osd_create(mpctx->global);
-
-    if (opts->force_vo) {
-        opts->fixed_vo = 1;
-        mpctx->video_out = init_best_video_out(mpctx->global, mpctx->input,
-                                               mpctx->encode_lavc_ctx);
-        if (!mpctx->video_out) {
-            MP_FATAL(mpctx, "Error opening/initializing "
-                    "the selected video_out (-vo) device.\n");
-            exit_player(mpctx, EXIT_ERROR);
-        }
-        mpctx->mouse_cursor_visible = true;
-        mpctx->initialized_flags |= INITIALIZED_VO;
-    }
-
-#if HAVE_LUA
-    // Lua user scripts can call arbitrary functions. Load them at a point
-    // where this is safe.
-    mp_lua_init(mpctx);
-#endif
-
-    if (opts->shuffle)
-        playlist_shuffle(mpctx->playlist);
-
-    if (opts->merge_files)
-        merge_playlist_files(mpctx->playlist);
-
-    mpctx->playlist->current = mp_check_playlist_resume(mpctx, mpctx->playlist);
-    if (!mpctx->playlist->current)
-        mpctx->playlist->current = mpctx->playlist->first;
+    if (mp_initialize(mpctx) < 0)
+        exit_player(mpctx, EXIT_ERROR);
 
     mp_play_files(mpctx);
 
@@ -462,12 +535,3 @@ static int mpv_main(int argc, char *argv[])
 
     return 1;
 }
-
-int main(int argc, char *argv[])
-{
-#if HAVE_COCOA
-    return cocoa_main(mpv_main, argc, argv);
-#else
-    return mpv_main(argc, argv);
-#endif
-}
