@@ -24,15 +24,20 @@
 #include <string.h>
 #include <unistd.h>
 #include <string.h>
+#include <libgen.h>
+#include <errno.h>
+#include <stdint.h>
+
+#include <dvdread/dvd_reader.h>
+#include <dvdread/ifo_types.h>
+#include <dvdread/ifo_read.h>
+#include <dvdread/nav_read.h>
 
 #include "config.h"
 #include "talloc.h"
 #include "common/common.h"
 #include "common/msg.h"
 
-#include <libgen.h>
-#include <errno.h>
-
 #define FIRST_AC3_AID 128
 #define FIRST_DTS_AID 136
 #define FIRST_MPG_AID 0
@@ -40,14 +45,11 @@
 
 #include "stream.h"
 #include "options/m_option.h"
+#include "options/options.h"
 
-#include "stream_dvd.h"
 #include "stream_dvd_common.h"
 
-static char* dvd_device_current;
-int dvd_angle=1;
-
-#define	LIBDVDREAD_VERSION(maj,min,micro)	((maj)*10000 + (min)*100 + (micro))
+#define LIBDVDREAD_VERSION(maj,min,micro)       ((maj)*10000 + (min)*100 + (micro))
 /*
  * Try to autodetect the libdvd-0.9.0 library
  * (0.9.0 removed the <dvdread/dvd_udf.h> header, and moved the two defines
@@ -56,54 +58,67 @@ int dvd_angle=1;
  */
 #ifndef DVDREAD_VERSION
 #if defined(DVD_VIDEO_LB_LEN) && defined(MAX_UDF_FILE_NAME_LEN)
-#define	DVDREAD_VERSION	LIBDVDREAD_VERSION(0,9,0)
+#define DVDREAD_VERSION LIBDVDREAD_VERSION(0,9,0)
 #else
-#define	DVDREAD_VERSION	LIBDVDREAD_VERSION(0,8,0)
+#define DVDREAD_VERSION LIBDVDREAD_VERSION(0,8,0)
 #endif
 #endif
 
+typedef struct {
+    int id; // 0 - 31 mpeg; 128 - 159 ac3; 160 - 191 pcm
+    int language;
+    int type;
+    int channels;
+} stream_language_t;
+
+typedef struct {
+  dvd_reader_t *dvd;
+  dvd_file_t *title;
+  ifo_handle_t *vmg_file;
+  tt_srpt_t *tt_srpt;
+  ifo_handle_t *vts_file;
+  vts_ptt_srpt_t *vts_ptt_srpt;
+  pgc_t *cur_pgc;
+//
+  int cur_title;
+  int cur_cell;
+  int last_cell;
+  int cur_pack;
+  int cell_last_pack;
+  int cur_pgc_idx;
+// Navi:
+  int packs_left;
+  dsi_t dsi_pack;
+  int angle_seek;
+  unsigned int *cell_times_table;
+// audio datas
+  int nr_of_channels;
+  stream_language_t audio_streams[32];
+// subtitles
+  int nr_of_subtitles;
+  stream_language_t subtitles[32];
+
+  int dvd_angle;
+  char *dvd_device_current;
+  int dvd_speed;
+  int dvd_title;
+
+  int cfg_title;
+  char *cfg_device;
+} dvd_priv_t;
 
 static const dvd_priv_t stream_priv_dflts = {
-  .cfg_title = 1,
+  .cfg_title = 0,
 };
 
 #define OPT_BASE_STRUCT dvd_priv_t
 /// URL definition
 static const m_option_t stream_opts_fields[] = {
-    OPT_INTRANGE("title", cfg_title, 0, 1, 99),
+    OPT_INTRANGE("title", cfg_title, 0, 0, 99),
     OPT_STRING("device", cfg_device, 0),
     {0}
 };
 
-int dvd_chapter_from_cell(dvd_priv_t* dvd,int title,int cell)
-{
-  pgc_t * cur_pgc;
-  ptt_info_t* ptt;
-  int chapter = cell;
-  int pgc_id,pgn;
-  if(title < 0 || cell < 0){
-    return 0;
-  }
-  /* for most DVD's chapter == cell */
-  /* but there are more complecated cases... */
-  if(chapter >= dvd->vmg_file->tt_srpt->title[title].nr_of_ptts) {
-    chapter = dvd->vmg_file->tt_srpt->title[title].nr_of_ptts-1;
-  }
-  title = dvd->tt_srpt->title[title].vts_ttn-1;
-  ptt = dvd->vts_file->vts_ptt_srpt->title[title].ptt;
-  while(chapter >= 0) {
-    pgc_id = ptt[chapter].pgcn;
-    pgn = ptt[chapter].pgn;
-    cur_pgc = dvd->vts_file->vts_pgcit->pgci_srp[pgc_id-1].pgc;
-    if(cell >= cur_pgc->program_map[pgn-1]-1) {
-      return chapter;
-    }
-    --chapter;
-  }
-  /* didn't find a chapter ??? */
-  return chapter;
-}
-
 static int dvd_lang_from_aid(stream_t *stream, int id) {
   dvd_priv_t *d;
   int i;
@@ -117,25 +132,7 @@ static int dvd_lang_from_aid(stream_t *stream, int id) {
   return 0;
 }
 
-int dvd_aid_from_lang(stream_t *stream, char **lang) {
-  dvd_priv_t *d=stream->priv;
-  int code,i;
-  for (int n = 0; lang[n]; n++) {
-      code = lang[n][1] | (lang[n][0] << 8);
-      for(i=0;i<d->nr_of_channels;i++) {
-        if(d->audio_streams[i].language==code) {
-          MP_INFO(stream, "Selected DVD audio channel: %d language: %c%c\n",
-          d->audio_streams[i].id, lang[n][0], lang[n][1]);
-          return d->audio_streams[i].id;
-        }
-        //printf("%X != %X  (%c%c)\n",code,d->audio_streams[i].language,lang[0],lang[1]);
-      }
-  }
-  MP_WARN(stream, "No matching DVD audio language found!\n");
-  return -1;
-}
-
-int dvd_number_of_subs(stream_t *stream) {
+static int dvd_number_of_subs(stream_t *stream) {
   int i;
   int maxid = -1;
   dvd_priv_t *d;
@@ -158,22 +155,6 @@ static int dvd_lang_from_sid(stream_t *stream, int id) {
   return 0;
 }
 
-int dvd_sid_from_lang(stream_t *stream, char **lang) {
-  dvd_priv_t *d=stream->priv;
-  int code,i;
-  for (int n = 0; lang[n]; n++) {
-    code = lang[n][1] | (lang[n][0] << 8);
-    for(i=0;i<d->nr_of_subtitles;i++) {
-      if(d->subtitles[i].language==code) {
-        MP_INFO(stream, "Selected DVD subtitle channel: %d language: %c%c\n", i, lang[n][0], lang[n][1]);
-        return d->subtitles[i].id;
-      }
-    }
-  }
-  MP_WARN(stream, "No matching DVD subtitle language found!\n");
-  return -1;
-}
-
 static int dvd_next_cell(stream_t *stream, dvd_priv_t *d) {
   int next_cell=d->cur_cell;
 
@@ -191,7 +172,7 @@ static int dvd_next_cell(stream_t *stream, dvd_priv_t *d) {
   if(next_cell>=d->last_cell)
     return -1; // EOF
   if(d->cur_pgc->cell_playback[next_cell].block_type == BLOCK_TYPE_ANGLE_BLOCK ) {
-    next_cell+=dvd_angle-1;
+    next_cell+=d->dvd_angle-1;
     if(next_cell>=d->last_cell)
       return -1; // EOF
   }
@@ -266,12 +247,12 @@ read_next:
 
       if(d->angle_seek) {
         int i,skip=0;
-        for(i=0;i<9;i++)	// check if all values zero:
+        for(i=0;i<9;i++)        // check if all values zero:
           if((skip=d->dsi_pack.sml_agli.data[i].address)!=0) break;
         if(skip && skip!=0x7fffffff) {
           // sml_agli table has valid data (at least one non-zero):
          d->cur_pack=d->dsi_pack.dsi_gi.nv_pck_lbn+
-         d->dsi_pack.sml_agli.data[dvd_angle-1].address;
+         d->dsi_pack.sml_agli.data[d->dvd_angle-1].address;
          d->angle_seek=0;
          d->cur_pack--;
          MP_VERBOSE(stream, "Angle-seek synced using sml_agli map!  new_lba=0x%X  \n",d->cur_pack);
@@ -305,38 +286,6 @@ read_next:
   return d->cur_pack-1;
 }
 
-static void dvd_seek(stream_t *stream, dvd_priv_t *d, int pos)
-{
-  d->packs_left=-1;
-  d->cur_pack=pos;
-
-  // check if we stay in current cell (speedup things, and avoid angle skip)
-  if(d->cur_pack>d->cell_last_pack ||
-     d->cur_pack<d->cur_pgc->cell_playback[ d->cur_cell ].first_sector) {
-
-    // ok, cell change, find the right cell!
-    cell_playback_t *cell;
-    for(d->cur_cell=0; d->cur_cell < d->cur_pgc->nr_of_cells; d->cur_cell++) {
-      cell = &(d->cur_pgc->cell_playback[d->cur_cell]);
-      if(cell->block_type == BLOCK_TYPE_ANGLE_BLOCK && cell->block_mode != BLOCK_MODE_FIRST_CELL)
-        continue;
-      d->cell_last_pack=cell->last_sector;
-      if(d->cur_pack<cell->first_sector) {
-        d->cur_pack=cell->first_sector;
-        break;
-      }
-      if(d->cur_pack<=d->cell_last_pack) break; // ok, we find it! :)
-    }
-  }
-
-  MP_VERBOSE(stream, "DVD Seek! lba=0x%X  cell=%d  packs: 0x%X-0x%X  \n",
-    d->cur_pack,d->cur_cell,d->cur_pgc->cell_playback[ d->cur_cell ].first_sector,d->cell_last_pack);
-
-  // if we're in interleaved multi-angle cell, find the right angle chain!
-  // (read Navi block, and use the seamless angle jump table)
-  d->angle_seek=1;
-}
-
 static int fill_buffer(stream_t *s, char *buf, int len)
 {
   int64_t pos;
@@ -345,23 +294,17 @@ static int fill_buffer(stream_t *s, char *buf, int len)
   pos = dvd_read_sector(s, s->priv, buf);
   if (pos < 0)
     return -1;
-  // dvd_read_sector() sometimes internally skips disk-level blocks
-  s->pos = 2048*(pos - 1);
   return 2048; // full sector
 }
 
-static int seek(stream_t *s, int64_t newpos) {
-  dvd_seek(s, s->priv,newpos/2048);
-  return 1;
-}
-
 static void stream_dvd_close(stream_t *s) {
   dvd_priv_t *d = s->priv;
   ifoClose(d->vts_file);
   ifoClose(d->vmg_file);
   DVDCloseFile(d->title);
   DVDClose(d->dvd);
-  dvd_set_speed(s,dvd_device_current, -1); /* -1 => restore default */
+  if (d->dvd_speed)
+    dvd_set_speed(s,d->dvd_device_current, -1); /* -1 => restore default */
 }
 
 static int mp_get_titleset_length(ifo_handle_t *vts_file, tt_srpt_t *tt_srpt, int title_no)
@@ -384,32 +327,6 @@ static int mp_get_titleset_length(ifo_handle_t *vts_file, tt_srpt_t *tt_srpt, in
 }
 
 
-static int mp_describe_titleset(stream_t *stream, dvd_reader_t *dvd, tt_srpt_t *tt_srpt, int vts_no)
-{
-    ifo_handle_t *vts_file;
-    int title_no, msec=0;
-
-    vts_file = ifoOpen(dvd, vts_no);
-    if(!vts_file)
-        return 0;
-
-    if(!vts_file->vtsi_mat || !vts_file->vts_pgcit)
-    {
-        ifoClose(vts_file);
-        return 0;
-    }
-
-    for(title_no = 0; title_no < tt_srpt->nr_of_srpts; title_no++)
-    {
-        if (tt_srpt->title[title_no].title_set_nr != vts_no)
-            continue;
-        msec = mp_get_titleset_length(vts_file, tt_srpt, title_no);
-        MP_SMODE(stream, "ID_DVD_TITLE_%d_LENGTH=%d.%03d\n", title_no + 1, msec / 1000, msec % 1000);
-    }
-    ifoClose(vts_file);
-    return 1;
-}
-
 static int get_num_chapter(ifo_handle_t *vts_file, tt_srpt_t *tt_srpt, int title_no)
 {
     if(!vts_file || !tt_srpt)
@@ -427,49 +344,6 @@ static int get_num_chapter(ifo_handle_t *vts_file, tt_srpt_t *tt_srpt, int title
     return vts_file->vts_ptt_srpt->title[title_no].nr_of_ptts;
 }
 
-static int seek_to_chapter(stream_t *stream, ifo_handle_t *vts_file, tt_srpt_t *tt_srpt, int title_no, int chapter)
-{
-    dvd_priv_t *d = stream->priv;
-    ptt_info_t ptt;
-    pgc_t *pgc;
-    int64_t pos;
-
-    if(!vts_file || !tt_srpt)
-       return 0;
-
-    if(title_no < 0 || title_no >= tt_srpt->nr_of_srpts)
-       return 0;
-
-    // map global title to vts title
-    title_no = tt_srpt->title[title_no].vts_ttn - 1;
-
-    if(title_no < 0 || title_no >= vts_file->vts_ptt_srpt->nr_of_srpts)
-       return 0;
-
-    if(chapter < 0 || chapter > vts_file->vts_ptt_srpt->title[title_no].nr_of_ptts-1) //no such chapter
-       return 0;
-
-    ptt = vts_file->vts_ptt_srpt->title[title_no].ptt[chapter];
-    pgc = vts_file->vts_pgcit->pgci_srp[ptt.pgcn-1].pgc;
-
-    d->cur_cell = pgc->program_map[ptt.pgn - 1] - 1;
-    if(pgc->cell_playback[d->cur_cell].block_type == BLOCK_TYPE_ANGLE_BLOCK)
-       d->cur_cell += dvd_angle-1;
-    d->cur_pack       = pgc->cell_playback[d->cur_cell].first_sector;
-    d->cell_last_pack = pgc->cell_playback[d->cur_cell].last_sector;
-
-    d->packs_left     = -1;
-    d->angle_seek     = 0;
-
-    pos = (int64_t) d->cur_pack * 2048;
-    stream_seek(stream, pos);
-
-    MP_VERBOSE(stream, "\r\nSTREAM_DVD, seeked to chapter: %d, cell: %u, pos: %"PRIu64"\n",
-        chapter, d->cur_pack, pos);
-
-    return chapter;
-}
-
 // p: in=chapter number, out=PTS
 static int get_chapter_time(ifo_handle_t *vts_file, tt_srpt_t *tt_srpt, int title_no, double *p)
 {
@@ -543,6 +417,44 @@ static double dvd_get_current_time(stream_t *stream, int cell)
     return (double)tm/1000.0;
 }
 
+static void dvd_seek(stream_t *stream, dvd_priv_t *d, int pos)
+{
+  d->packs_left=-1;
+  d->cur_pack=pos;
+
+  // check if we stay in current cell (speedup things, and avoid angle skip)
+  if(d->cur_pack>d->cell_last_pack ||
+     d->cur_pack<d->cur_pgc->cell_playback[ d->cur_cell ].first_sector) {
+
+    // ok, cell change, find the right cell!
+    cell_playback_t *cell;
+    for(d->cur_cell=0; d->cur_cell < d->cur_pgc->nr_of_cells; d->cur_cell++) {
+      cell = &(d->cur_pgc->cell_playback[d->cur_cell]);
+      if(cell->block_type == BLOCK_TYPE_ANGLE_BLOCK && cell->block_mode != BLOCK_MODE_FIRST_CELL)
+        continue;
+      d->cell_last_pack=cell->last_sector;
+      if(d->cur_pack<cell->first_sector) {
+        d->cur_pack=cell->first_sector;
+        break;
+      }
+      if(d->cur_pack<=d->cell_last_pack) break; // ok, we find it! :)
+    }
+  }
+
+  MP_VERBOSE(stream, "DVD Seek! lba=0x%X  cell=%d  packs: 0x%X-0x%X  \n",
+    d->cur_pack,d->cur_cell,d->cur_pgc->cell_playback[ d->cur_cell ].first_sector,d->cell_last_pack);
+
+  // if we're in interleaved multi-angle cell, find the right angle chain!
+  // (read Navi block, and use the seamless angle jump table)
+  d->angle_seek=1;
+}
+
+static int do_seek(stream_t *s, int64_t newpos) {
+  stream_drop_buffers(s);
+  dvd_seek(s, s->priv,newpos/2048);
+  return 1;
+}
+
 static int dvd_seek_to_time(stream_t *stream, ifo_handle_t *vts_file, double sec)
 {
     unsigned int i, j, k, timeunit, ac_time, tmap_sector=0, cell_sector=0, vobu_sector=0;
@@ -555,7 +467,7 @@ static int dvd_seek_to_time(stream_t *stream, ifo_handle_t *vts_file, double sec
     if(!vts_file->vts_tmapt || sec < 0)
         return 0;
 
-    duration = (double) mp_get_titleset_length(d->vts_file, d->tt_srpt, d->cur_title-1) / 1000.0f;
+    duration = (double) mp_get_titleset_length(d->vts_file, d->tt_srpt, d->cur_title) / 1000.0f;
     if(sec > duration)
       return 0;
 
@@ -576,23 +488,27 @@ static int dvd_seek_to_time(stream_t *stream, ifo_handle_t *vts_file, double sec
     }
 
     pos = ((int64_t)cell_sector)<<11;
-    stream_seek(stream, pos);
+    do_seek(stream, pos);
     do {
-      stream_skip(stream, 2048);
+      char buf[2048];
+      if (dvd_read_sector(stream, stream->priv, buf) < 0) // skip
+          break;
       t = mp_dvdtimetomsec(&d->dsi_pack.dsi_gi.c_eltm);
     } while(!t);
     tm = dvd_get_current_time(stream, -1);
 
     pos = ((int64_t)tmap_sector)<<11;
-    stream_seek(stream, pos);
+    do_seek(stream, pos);
     //now get current time in terms of the cell+cell time offset
     memset(&d->dsi_pack.dsi_gi.c_eltm, 0, sizeof(dvd_time_t));
     while(tm <= sec) {
-        if(!stream_skip(stream, 2048))
-          break;
+        char buf[2048];
+        if (dvd_read_sector(stream, stream->priv, buf) < 0) // skip
+            break;
+        pos += 2048;
         tm = dvd_get_current_time(stream, -1);
     };
-    tmap_sector = stream->pos >> 11;
+    tmap_sector = pos >> 11;
 
     //search closest VOBU sector
     k=(vts_file->vts_vobu_admap->last_byte + 1 - VOBU_ADMAP_SIZE)/4; //entries in the vobu admap
@@ -602,7 +518,7 @@ static int dvd_seek_to_time(stream_t *stream, ifo_handle_t *vts_file, double sec
     }
     vobu_sector = vts_file->vts_vobu_admap->vobu_start_sectors[i-1];
     pos = ((int64_t)vobu_sector) << 11;
-    stream_seek(stream, pos);
+    do_seek(stream, pos);
 
     return 1;
 }
@@ -614,7 +530,7 @@ static int control(stream_t *stream,int cmd,void* arg)
     {
         case STREAM_CTRL_GET_TIME_LENGTH:
         {
-            *((double *)arg) = (double) mp_get_titleset_length(d->vts_file, d->tt_srpt, d->cur_title-1)/1000.0;
+            *((double *)arg) = (double) mp_get_titleset_length(d->vts_file, d->tt_srpt, d->cur_title)/1000.0;
             return 1;
         }
         case STREAM_CTRL_GET_START_TIME:
@@ -630,7 +546,7 @@ static int control(stream_t *stream,int cmd,void* arg)
         case STREAM_CTRL_GET_NUM_CHAPTERS:
         {
             int r;
-            r = get_num_chapter(d->vts_file, d->tt_srpt, d->cur_title-1);
+            r = get_num_chapter(d->vts_file, d->tt_srpt, d->cur_title);
             if(! r) return STREAM_UNSUPPORTED;
             *((unsigned int *)arg) = r;
             return 1;
@@ -638,26 +554,13 @@ static int control(stream_t *stream,int cmd,void* arg)
         case STREAM_CTRL_GET_CHAPTER_TIME:
         {
             int r;
-            r = get_chapter_time(d->vts_file, d->tt_srpt, d->cur_title-1, (double *)arg);
+            r = get_chapter_time(d->vts_file, d->tt_srpt, d->cur_title, (double *)arg);
             if(! r) return STREAM_UNSUPPORTED;
             return 1;
         }
-        case STREAM_CTRL_SEEK_TO_CHAPTER:
-        {
-            int r;
-            r = seek_to_chapter(stream, d->vts_file, d->tt_srpt, d->cur_title-1, *((unsigned int *)arg));
-            if(! r) return STREAM_UNSUPPORTED;
-
-            return 1;
-        }
         case STREAM_CTRL_GET_CURRENT_TITLE:
         {
-            *((unsigned int *)arg) = d->cur_title - 1;
-            return 1;
-        }
-        case STREAM_CTRL_GET_CURRENT_CHAPTER:
-        {
-            *((unsigned int *)arg) = dvd_chapter_from_cell(d, d->cur_title-1, d->cur_cell);
+            *((unsigned int *)arg) = d->cur_title;
             return 1;
         }
         case STREAM_CTRL_GET_CURRENT_TIME:
@@ -683,20 +586,20 @@ static int control(stream_t *stream,int cmd,void* arg)
         }
         case STREAM_CTRL_GET_NUM_ANGLES:
         {
-            *((int *)arg) = d->vmg_file->tt_srpt->title[dvd_title].nr_of_angles;
+            *((int *)arg) = d->vmg_file->tt_srpt->title[d->dvd_title].nr_of_angles;
             return 1;
         }
         case STREAM_CTRL_GET_ANGLE:
         {
-            *((int *)arg) = dvd_angle;
+            *((int *)arg) = d->dvd_angle;
             return 1;
         }
         case STREAM_CTRL_SET_ANGLE:
         {
             int ang = *((int *)arg);
-            if(ang>d->vmg_file->tt_srpt->title[dvd_title].nr_of_angles || ang<=0)
+            if(ang>d->vmg_file->tt_srpt->title[d->dvd_title].nr_of_angles || ang<=0)
                 break;
-            dvd_angle = ang;
+            d->dvd_angle = ang;
             d->angle_seek = 1;
             return 1;
         }
@@ -727,18 +630,31 @@ static int control(stream_t *stream,int cmd,void* arg)
             memcpy(req->palette, d->cur_pgc->palette, sizeof(req->palette));
             return STREAM_OK;
         }
+        case STREAM_CTRL_GET_DISC_NAME:
+        {
+            char buffer[128];
+            if (DVDUDFVolumeInfo(d->dvd, buffer, sizeof(buffer), NULL, 0) < 0 &&
+                DVDISOVolumeInfo(d->dvd, buffer, sizeof(buffer), NULL, 0) < 0)
+                break;
+            if (!buffer[0])
+                break;
+            *(char**)arg = talloc_strdup(NULL, buffer);
+            return STREAM_OK;
+        }
     }
     return STREAM_UNSUPPORTED;
 }
 
 
-static int open_s(stream_t *stream, int mode)
+static int open_s(stream_t *stream)
 {
   int k;
   dvd_priv_t *d = stream->priv;
 
+  d->dvd_angle = stream->opts->dvd_angle;
+
   MP_VERBOSE(stream, "URL: %s\n", stream->url);
-  dvd_title = d->cfg_title;
+  d->dvd_title = d->cfg_title + 1;
   if(1){
     //int ret,ret2;
     int ttn,pgc_id,pgn;
@@ -751,18 +667,19 @@ static int open_s(stream_t *stream, int mode)
     /**
      * Open the disc.
      */
-    if(d->cfg_device)
-      dvd_device_current = d->cfg_device;
-    else if(dvd_device)
-      dvd_device_current = dvd_device;
+    if(d->cfg_device && d->cfg_device[0])
+      d->dvd_device_current = d->cfg_device;
+    else if(stream->opts->dvd_device && stream->opts->dvd_device[0])
+      d->dvd_device_current = talloc_strdup(stream, stream->opts->dvd_device);
     else
-      dvd_device_current = DEFAULT_DVD_DEVICE;
-    dvd_set_speed(stream,dvd_device_current, dvd_speed);
+      d->dvd_device_current = DEFAULT_DVD_DEVICE;
+    d->dvd_speed = stream->opts->dvd_speed;
+    dvd_set_speed(stream,d->dvd_device_current, d->dvd_speed);
 #if defined(__APPLE__) || defined(__DARWIN__)
     /* Dynamic DVD drive selection on Darwin */
-    if(!strcmp(dvd_device_current, "/dev/rdiskN")) {
+    if(!strcmp(d->dvd_device_current, "/dev/rdiskN")) {
       int i;
-      size_t len = strlen(dvd_device_current)+1;
+      size_t len = strlen(d->dvd_device_current)+1;
       char *temp_device = malloc(len);
 
       for (i = 1; i < 10; i++) {
@@ -772,7 +689,7 @@ static int open_s(stream_t *stream, int mode)
           MP_ERR(stream, "Couldn't open DVD device: %s (%s)\n",temp_device, strerror(errno));
         } else {
 #if DVDREAD_VERSION <= LIBDVDREAD_VERSION(0,9,4)
-          dvd_file_t *dvdfile = DVDOpenFile(dvd,dvd_title,DVD_READ_INFO_FILE);
+          dvd_file_t *dvdfile = DVDOpenFile(dvd,d->dvd_title,DVD_READ_INFO_FILE);
           if(!dvdfile) {
             MP_ERR(stream, "Couldn't open DVD device: %s (%s)\n",temp_device, strerror(errno));
             DVDClose(dvd);
@@ -791,9 +708,9 @@ static int open_s(stream_t *stream, int mode)
     } else
 #endif /* defined(__APPLE__) || defined(__DARWIN__) */
     {
-        dvd = DVDOpen(dvd_device_current);
+        dvd = DVDOpen(d->dvd_device_current);
         if(!dvd) {
-          MP_ERR(stream, "Couldn't open DVD device: %s (%s)\n",dvd_device_current, strerror(errno));
+          MP_ERR(stream, "Couldn't open DVD device: %s (%s)\n",d->dvd_device_current, strerror(errno));
           return STREAM_UNSUPPORTED;
         }
     }
@@ -811,70 +728,41 @@ static int open_s(stream_t *stream, int mode)
       return STREAM_UNSUPPORTED;
     }
     tt_srpt = vmg_file->tt_srpt;
-    if (mp_msg_test(stream->log, MSGL_SMODE))
-    {
-      int title_no; ///< title number
-      MP_SMODE(stream, "ID_DVD_TITLES=%d\n", tt_srpt->nr_of_srpts);
-      for (title_no = 0; title_no < tt_srpt->nr_of_srpts; title_no++)
-      {
-        MP_SMODE(stream, "ID_DVD_TITLE_%d_CHAPTERS=%d\n", title_no + 1, tt_srpt->title[title_no].nr_of_ptts);
-        MP_SMODE(stream, "ID_DVD_TITLE_%d_ANGLES=%d\n", title_no + 1, tt_srpt->title[title_no].nr_of_angles);
-      }
-    }
-    if (mp_msg_test(stream->log, MSGL_SMODE))
-    {
-      char volid[32];
-      unsigned char discid [16]; ///< disk ID, a 128 bit MD5 sum
-      int vts_no;   ///< video title set number
-      for (vts_no = 1; vts_no <= vmg_file->vts_atrt->nr_of_vtss; vts_no++)
-        mp_describe_titleset(stream, dvd, tt_srpt, vts_no);
-      if (DVDDiscID(dvd, discid) >= 0)
-      {
-        int i;
-        MP_SMODE(stream, "ID_DVD_DISC_ID=");
-        for (i = 0; i < 16; i ++)
-          MP_SMODE(stream, "%02X", discid[i]);
-        MP_SMODE(stream, "\n");
-      }
-      if (DVDUDFVolumeInfo(dvd, volid, sizeof(volid), NULL, 0) >= 0 || DVDISOVolumeInfo(dvd, volid, sizeof(volid), NULL, 0) >= 0)
-        MP_SMODE(stream, "ID_DVD_VOLUME_ID=%s\n", volid);
-    }
     /**
      * Make sure our title number is valid.
      */
     MP_INFO(stream, "There are %d titles on this DVD.\n", tt_srpt->nr_of_srpts );
-    if(dvd_title < 1 || dvd_title > tt_srpt->nr_of_srpts) {
-      MP_ERR(stream, "Invalid DVD title number: %d\n", dvd_title);
+    if(d->dvd_title < 1 || d->dvd_title > tt_srpt->nr_of_srpts) {
+      MP_ERR(stream, "Invalid DVD title number: %d\n", d->dvd_title);
       ifoClose( vmg_file );
       DVDClose( dvd );
       return STREAM_UNSUPPORTED;
     }
-    MP_SMODE(stream, "ID_DVD_CURRENT_TITLE=%d\n", dvd_title);
-    --dvd_title; // remap 1.. -> 0..
+    --(d->dvd_title); // remap 1.. -> 0..
     /**
      * Make sure the angle number is valid for this title.
      */
-    MP_INFO(stream, "There are %d angles in this DVD title.\n", tt_srpt->title[dvd_title].nr_of_angles);
-    if(dvd_angle<1 || dvd_angle>tt_srpt->title[dvd_title].nr_of_angles) {
-      MP_ERR(stream, "Invalid DVD angle number: %d\n", dvd_angle);
+    MP_INFO(stream, "There are %d angles in this DVD title.\n", tt_srpt->title[d->dvd_title].nr_of_angles);
+    if(d->dvd_angle<1 || d->dvd_angle>tt_srpt->title[d->dvd_title].nr_of_angles) {
+      MP_ERR(stream, "Invalid DVD angle number: %d\n", d->dvd_angle);
       goto fail;
     }
 
-    ttn = tt_srpt->title[dvd_title].vts_ttn - 1;
+    ttn = tt_srpt->title[d->dvd_title].vts_ttn - 1;
     /**
      * Load the VTS information for the title set our title is in.
      */
-    vts_file = ifoOpen( dvd, tt_srpt->title[dvd_title].title_set_nr );
+    vts_file = ifoOpen( dvd, tt_srpt->title[d->dvd_title].title_set_nr );
     if(!vts_file) {
-      MP_ERR(stream, "Cannot open the IFO file for DVD title %d.\n", tt_srpt->title[dvd_title].title_set_nr );
+      MP_ERR(stream, "Cannot open the IFO file for DVD title %d.\n", tt_srpt->title[d->dvd_title].title_set_nr );
       goto fail;
     }
     /**
      * We've got enough info, time to open the title set data.
      */
-    title = DVDOpenFile(dvd, tt_srpt->title[dvd_title].title_set_nr, DVD_READ_TITLE_VOBS);
+    title = DVDOpenFile(dvd, tt_srpt->title[d->dvd_title].title_set_nr, DVD_READ_TITLE_VOBS);
     if(!title) {
-      MP_ERR(stream, "Cannot open title VOBS (VTS_%02d_1.VOB).\n", tt_srpt->title[dvd_title].title_set_nr);
+      MP_ERR(stream, "Cannot open title VOBS (VTS_%02d_1.VOB).\n", tt_srpt->title[d->dvd_title].title_set_nr);
       ifoClose( vts_file );
       goto fail;
     }
@@ -886,7 +774,7 @@ static int open_s(stream_t *stream, int mode)
     d->vmg_file=vmg_file;
     d->tt_srpt=tt_srpt;
     d->vts_file=vts_file;
-    d->cur_title = dvd_title+1;
+    d->cur_title = d->dvd_title;
 
     pgc = vts_file->vts_pgcit ? vts_file->vts_pgcit->pgci_srp[ttn].pgc : NULL;
     /**
@@ -940,9 +828,6 @@ static int open_s(stream_t *stream, int mode)
              tmp,
              audio_stream->id
            );
-           MP_SMODE(stream, "ID_AUDIO_ID=%d\n", audio_stream->id);
-           if(language && tmp[0])
-             MP_SMODE(stream, "ID_AID_%d_LANG=%s\n", audio_stream->id, tmp);
 
            d->nr_of_channels++;
          }
@@ -980,9 +865,6 @@ static int open_s(stream_t *stream, int mode)
           sub_stream->id = pgc->subp_control[i] >> 8 & 31;
 
         MP_INFO(stream, "subtitle ( sid ): %d language: %s\n", sub_stream->id, tmp);
-        MP_SMODE(stream, "ID_SUBTITLE_ID=%d\n", sub_stream->id);
-        if(language && tmp[0])
-          MP_SMODE(stream, "ID_SID_%d_LANG=%s\n", sub_stream->id, tmp);
         d->nr_of_subtitles++;
       }
       MP_INFO(stream, "number of subtitles on disk: %d\n",d->nr_of_subtitles);
@@ -1002,7 +884,7 @@ static int open_s(stream_t *stream, int mode)
       d->last_cell=d->cur_pgc->nr_of_cells;
 
     if(d->cur_pgc->cell_playback[d->cur_cell].block_type == BLOCK_TYPE_ANGLE_BLOCK )
-      d->cur_cell+=dvd_angle-1;
+      d->cur_cell+=d->dvd_angle-1;
     d->cur_pack = d->cur_pgc->cell_playback[ d->cur_cell ].first_sector;
     d->cell_last_pack=d->cur_pgc->cell_playback[ d->cur_cell ].last_sector;
     MP_VERBOSE(stream, "DVD start cell: %d  pack: 0x%X-0x%X  \n",d->cur_cell,d->cur_pack,d->cell_last_pack);
@@ -1013,18 +895,15 @@ static int open_s(stream_t *stream, int mode)
       return STREAM_UNSUPPORTED;
     for(k=0; k<d->cur_pgc->nr_of_cells; k++)
       d->cell_times_table[k] = mp_dvdtimetomsec(&d->cur_pgc->cell_playback[k].playback_time);
-    list_chapters(stream, vts_file,tt_srpt,dvd_title);
+    list_chapters(stream, vts_file,tt_srpt,d->dvd_title);
 
     // ... (unimplemented)
     //    return NULL;
     stream->type = STREAMTYPE_DVD;
     stream->sector_size = 2048;
-    stream->flags = MP_STREAM_SEEK;
     stream->fill_buffer = fill_buffer;
-    stream->seek = seek;
     stream->control = control;
     stream->close = stream_dvd_close;
-    stream->start_pos = (int64_t)d->cur_pack*2048;
     stream->end_pos = (int64_t)(d->cur_pgc->cell_playback[d->last_cell-1].last_sector)*2048;
     MP_VERBOSE(stream, "DVD start=%d end=%d  \n",d->cur_pack,d->cur_pgc->cell_playback[d->last_cell-1].last_sector);
     stream->priv = (void*)d;
@@ -1039,7 +918,7 @@ fail:
   return STREAM_UNSUPPORTED;
 }
 
-static int ifo_stream_open (stream_t *stream, int mode)
+static int ifo_stream_open (stream_t *stream)
 {
     char* filename;
     dvd_priv_t *priv = talloc_ptrtype(stream, priv);
@@ -1062,25 +941,25 @@ static int ifo_stream_open (stream_t *stream, int mode)
     priv->cfg_device = talloc_strdup(NULL, dirname(stream->path));
     if(!strncasecmp(filename,"vts_",4))
     {
-        if(sscanf(filename+3, "_%02d_", &priv->cfg_title)!=1)
-            priv->cfg_title = 1;
+        if(sscanf(filename+3, "_%02d_", &priv->cfg_title)!=0)
+            priv->cfg_title = 0;
     }else
-        priv->cfg_title = 1;
+        priv->cfg_title = 0;
 
     free(filename);
-    stream->url=talloc_strdup(stream, "dvd://");
+    stream->url=talloc_strdup(stream, "dvdread://");
 
-    return open_s(stream, mode);
+    return open_s(stream);
 }
 
 const stream_info_t stream_info_dvd = {
   .name = "dvd",
   .open = open_s,
-  .protocols = (const char*[]){ "dvd", NULL },
+  .protocols = (const char*const[]){ "dvdread", NULL },
   .priv_size = sizeof(dvd_priv_t),
   .priv_defaults = &stream_priv_dflts,
   .options = stream_opts_fields,
-  .url_options = (const char*[]){
+  .url_options = (const char*const[]){
         "hostname=title",
         "filename=device",
         NULL
@@ -1090,5 +969,5 @@ const stream_info_t stream_info_dvd = {
 const stream_info_t stream_info_ifo = {
     .name = "ifo",
     .open = ifo_stream_open,
-    .protocols = (const char*[]){ "file", "", NULL },
+    .protocols = (const char*const[]){ "file", "", NULL },
 };
